/*
This file is part of the Project JsBuild

JsBuild - http://projects.stueckseln.de/jsbuild/
JsBuild is a project released under the MIT license.


Copyright (c) 2009 Thomas Lehmann

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package de.stueckseln.buildtools.jsbuild;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.text.DecimalFormat;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.mozilla.javascript.ErrorReporter;
import org.mozilla.javascript.EvaluatorException;

import com.yahoo.platform.yui.compressor.JavaScriptCompressor;

import de.stueckseln.buildtools.utils.Logger;
import de.stueckseln.buildtools.utils.PathTool;
import de.stueckseln.buildtools.utils.SimpleLogger;
import de.stueckseln.buildtools.utils.SimplePathBuilder;

import de.stueckseln.buildtools.utils.FileHelper;

/** This abstract class will compress configured modules.
  * The abstract method parseModules() has to be implemented in a derived class.
  * That method should parse/load a configuration and store it in a list of modules.
  * @see JsModule
  * 
  * @author Thomas Lehmann
  */
public abstract class AbstractJsBuilder {
	private static final String LINE = "--------------------------------------------------------------";
	
	public static final char PATHSEP = '/';  
	public static final String NAME_BUILT_POSTFIX = "-build.js";
	public static final String NAME_ORDERLIST_POSTFIX = "-includelist.txt";
	public static final String NAME_TMP_PRECOMPESS_POSTFIX = "-precompress.js";
	
	protected char sep = PATHSEP;
	
	protected String sourceDirectory = ".";
	protected String targetDirectory = ".";
	
	protected String descriptorFile;

	protected String builtPostfix;
	protected String builtOrderList;
	
	protected String prependString;
	protected String appendString;
	
	public boolean isCompressEnabled() {
		return compressEnabled;
	}
	public void setCompressEnabled(boolean compressEnabled) {
		this.compressEnabled = compressEnabled;
	}
	public boolean isShowCompressorWarnings() {
		return showCompressorWarnings;
	}
	public void setShowCompressorWarnings(boolean showCompressorWarnings) {
		this.showCompressorWarnings = showCompressorWarnings;
	}
	public boolean isShowInfos() {
		return showInfos;
	}
	public void setShowInfos(boolean showInfos) {
		this.showInfos = showInfos;
	}
	
	/** Returns the current value for parameter no-munge for YUI compressor
	  * @return true if parameter no-munge is set (to true)
	  */
	public boolean isYuiNoMunge() {
		return yuiNoMunge;
	}
	/** Set parameter no-munge for YUI compressor
	  * Unless of it's name this method should be set to true to compress code
	  * @param yuiNoMunge new value for no-munge 
	  */
	public void setYuiNoMunge(boolean yuiNoMunge) {
		this.yuiNoMunge = yuiNoMunge;
	}

	public boolean isYuiNoOpts() {
		return yuiNoOpts;
	}
	public void setYuiNoOpts(boolean yuiNoOpts) {
		this.yuiNoOpts = yuiNoOpts;
	}
	public boolean isYuiPreserveSemicolons() {
		return yuiPreserveSemicolons;
	}
	public void setYuiPreserveSemicolons(boolean yuiPreserveSemicolons) {
		this.yuiPreserveSemicolons = yuiPreserveSemicolons;
	}
	public int getYuiLinebreak() {
		return yuiLinebreak;
	}
	public void setYuiLinebreak(int yuiLinebreak) {
		this.yuiLinebreak = yuiLinebreak;
	}

	protected boolean compressEnabled;
	protected boolean showCompressorWarnings;
	protected boolean showInfos;

	protected boolean yuiNoMunge;
	protected boolean yuiNoOpts;
	protected boolean yuiPreserveSemicolons;
	protected int yuiLinebreak;
	
	protected boolean overwriteExisting;
	
	
	public boolean isOverwriteExisting() {
		return overwriteExisting;
	}
	public void setOverwriteExisting(boolean overwriteExisting) {
		this.overwriteExisting = overwriteExisting;
	}
	public String getSourceDirectory() {
		return sourceDirectory;
	}
	public void setSourceDirectory(String sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}
	public String getTargetDirectory() {
		return targetDirectory;
	}
	public void setTargetDirectory(String targetDirectory) {
		this.targetDirectory = targetDirectory;
	}
	public String getModulesFile() {
		return descriptorFile;
	}
	public void setModulesFile(String modulesFile) {
		this.descriptorFile = modulesFile;
	}
	public String getBuiltPostfix() {
		return builtPostfix;
	}
	public void setBuiltPostfix(String builtPostfix) {
		this.builtPostfix = builtPostfix;
	}
	public String getBuiltOrderList() {
		return builtOrderList;
	}
	public void setBuiltOrderList(String builtOrderList) {
		this.builtOrderList = builtOrderList;
	}
	public String getPrependString() {
		return prependString;
	}
	public void setPrependString(String prependString) {
		this.prependString = prependString;
	}
	public String getAppendString() {
		return appendString;
	}
	public void setAppendString(String appendString) {
		this.appendString = appendString;
	}
	
	protected Logger LOG;
	
	protected List<JsModule> modules;	
	
	/** Initialize this instance with a Logger
	  * @param logger the logger to use
	  */
	public AbstractJsBuilder(Logger logger) throws JsBuilderException {
		this();
		
		LOG = logger;		
	}
	
	/** Initialize this instance with System.out as default output for logger
	  */
	public AbstractJsBuilder() throws JsBuilderException {
		LOG = new SimpleLogger().setOutputStream(System.out).setLogLevel(Logger.LOGLEVEL_DEBUG);
		
		this.builtOrderList = NAME_ORDERLIST_POSTFIX;
		this.builtPostfix = NAME_BUILT_POSTFIX;
				
		setSourceDirectory(".");
		setTargetDirectory(".");
		
		this.descriptorFile = null;
		
		this.prependString = null;
		this.appendString = null;
		
		this.modules = new LinkedList<JsModule>();
	}
	
	abstract protected void parseModules() throws JsBuilderException;
	
	/** Process a given module
	  * 
	  * @param module a Module instance to compress
	  * @throws JsBuilderException
	  */
	protected void processModule(JsModule module) throws JsBuilderException {
		long overallInputSize = 0;
		long overallOutputSize = 0;
		int fileCount = 0;
		
		StringBuffer logFileList = new StringBuffer();
		SimplePathBuilder pb = new SimplePathBuilder(); // TODO move to <this>
		pb.init(this.targetDirectory);
		
		pb.append(module.getName());
		String buildFileName = pb.getPath() + this.builtPostfix;
		String orderListFileName = pb.getPath() + this.builtOrderList;
		
		File buildFile = new File(buildFileName);
		File orderListFile = new File(orderListFileName);
		
		if (buildFile.exists()) {
			if (!isOverwriteExisting()) {
				throw new JsBuilderFileException("Buildfile '"+ buildFile +"' exists.");
			} else {
				LOG.info("Output file '" + buildFile.getName() + "' exists, overwriting...");
				buildFile.delete();
			}
		}
		
		boolean createFileSuccessful = true;
		try {
			createFileSuccessful = buildFile.createNewFile();
		} catch (IOException ioe) {
			createFileSuccessful = false;
		}
		
		if (!createFileSuccessful){
			throw new JsBuilderFileException("Cannot create buildfile '"+buildFile+"'");
		}
				
		String inputString = null;
		StringBuffer outputBuffer = new StringBuffer();
		StringBuffer orderListBuffer = new StringBuffer();
		StringBuffer compressedBuffer = null; 
		
		Iterator<String> includeIt = module.getFiles().iterator();
		String includeFilename;
		File includeFile;
		
		// the following iterator wont start the while loop
		if (module.getFiles().size() > 0) {
			if (this.isCompressEnabled()){
				LOG.info("Compressing ...");
			} else {
				LOG.info("Appending files ...");
			}			
		}
				
		// iterate over all files 
		while(includeIt.hasNext()){
			fileCount++;
			
			includeFilename = includeIt.next();
			String orderPath = PathTool.relativePath(this.sourceDirectory, includeFilename);
			
			orderListBuffer.append(orderPath);
			orderListBuffer.append("\n");

			includeFile = new File(includeFilename);
			
			// logFileList contains all processed filenames to output 
			logFileList.append(includeFile.getName());
			logFileList.append(", ");
			

			// we'll first compress a file, then put that compressed line into the outbuffer
			
			// read input file
			try {
				inputString = FileHelper.readFileToString(includeFile);
			} catch (IOException e) {
				throw new JsBuilderException("Cannot read include file: " + e.getMessage());
			}
			
			overallInputSize += inputString.length();

			// master-switch
			if (this.isCompressEnabled()){
				// TODO test for bad conditions (null) ?
				
				try {
					compressedBuffer = compress(inputString, module.getName(), includeFile.getName());
				} catch(Exception e){
					throw new JsBuilderException("Cannot compress file '"+includeFile.getName()+"' in module '"+module.getName()+"': " + e.getMessage());
				} catch (Error e) {
					throw new JsBuilderException("Possible classpath error when compressing: " + e.getMessage());
				}
			
				outputBuffer.append(compressedBuffer);
			} else {
				outputBuffer.append(inputString);
			}
			
			outputBuffer.append("\n\n");
			
			//LOG.debug("adding ["+module.getName()+"]: " + orderPath);
		}

		overallOutputSize = outputBuffer.length();
		
		// add prepend string
		if (this.prependString != null) {
			outputBuffer.insert(0, prependString);
		}

		// add append string
		if (this.appendString != null) {
			outputBuffer.append(appendString);
		}
		
		// logFileList contains all processed filenames to output 
		if (logFileList.length() > 2) {
			logFileList.delete(logFileList.length()-2, logFileList.length()-0);
		}
		
		if (fileCount > 0) {
			logFileList.insert(0, " files [");
			logFileList.insert(0, fileCount);
			logFileList.insert(0, "Added ");
			logFileList.append("]");
		} else {
			logFileList.append("No files in package.");
		}
		
		LOG.info(logFileList.toString());

		boolean compress = (this.isCompressEnabled() && this.yuiNoMunge);
		
		LOG.info("Overall input files size " + Formatter.size(overallInputSize));
		LOG.info("Output file size " + (compress ? "(compressed)" : "(uncompressed)") + " " + Formatter.size(overallOutputSize));
		if (overallInputSize != 0 && compress) {
			LOG.info("Compression ratio " + new DecimalFormat("###.00").format((double)(100-((double)(overallOutputSize * 100) / overallInputSize))) + "%");
		}
			
		// write orderlist
		try {
			FileHelper.writeFile(orderListBuffer, orderListFile);
		} catch (Exception e) {
			throw new JsBuilderException("Cannot write to order list file: " + e.getMessage());
		}
		
		// write compressed file
		try {
			FileHelper.writeFile(outputBuffer, buildFile);
		} catch (Exception e) {
			throw new JsBuilderException("Cannot write to build file: " + e.getMessage());
		}
				
	}
	
	
	/** Call YUI
 	  * @param source StringBuffer with source code to be compressed
 	  * @param moduleName the module name of currently processed package (to display in info/warn/error messages)
 	  * @param fileName the file name of currently processed source file
 	  * @return StringBuffer with compressed code
 	  * @throws Exception if any error occures
 	  */
	private StringBuffer compress(String source, String moduleName, String fileName) throws Exception {
		final Logger LOGGER = this.LOG;
		final boolean showErrors = true; 
		final boolean showWarns = this.showCompressorWarnings; 
		final boolean showRunErrors = this.showCompressorWarnings; 
		
		// finals for following anonymous class
		final String mName = moduleName;
		final String fName = fileName;
		
		ErrorReporter reporter = new ErrorReporter(){
			
			private String log(String prefix, String msg, String sourceName, int line, String lineSource, int column){
				StringBuffer str = new StringBuffer();
				
				// build output log line
				
				str.append("[YUIC] [");
				str.append(fName);
				str.append(" in ");
				str.append(mName);
				str.append("] ");
				
				// log type (Warning, Error, Runtime)
				if (prefix != null) {
					str.append("[");
					str.append(prefix);
					str.append("] ");
				}
				
				if (line > 0) {
					str.append(line);
					str.append(":");
					str.append(column);
					str.append(" ");
				}
				str.append(msg);
				
				return str.toString();
			}
			
			public void error(String msg, String sourceName, int line, String lineSource, int column) {
				if (showErrors) {
					LOGGER.error(log("E", msg, sourceName, line, lineSource, column));
				}
			}
			
			public void warning(String msg, String sourceName, int line, String lineSource, int column){
				if (showWarns) {
					LOGGER.warning(log("W", msg, sourceName, line, lineSource, column));
				}
			}

			public EvaluatorException runtimeError(String msg, String sourceName, int line, String lineSource, int column){
				if (showRunErrors) {
					LOGGER.error(log("R", msg, sourceName, line, lineSource, column));
				}				
				return new EvaluatorException(msg);
			}

		}; // of new ErrorReporter()

		Reader compressReader = new BufferedReader(new StringReader(source));
		
		StringWriter writer = new StringWriter();	
		Writer compressWriter = new BufferedWriter(writer);
		
		try {
			JavaScriptCompressor compress = new JavaScriptCompressor(compressReader, reporter); 
			compress.compress(compressWriter, yuiLinebreak, yuiNoMunge, true, yuiPreserveSemicolons, yuiNoOpts);
			
		} catch (Exception e) {
			throw e;
			
		} finally {
			compressReader.close();
			compressWriter.close();
			
		}
		
		return writer.getBuffer();
	
	}
	
	/** Run build.
	  * Abstract parseModules() (see @see parseModules() ) will be called first
	  * If configured modules where found the compression task will be run
	  * 
	  * 
	  * @throws JsBuilderException
	  */
	public void build() throws JsBuilderException {
		parseModules();
		if (modules.size() == 0){
			LOG.info("No modules found.");
			return;
		}
	
		LOG.info("Building to '" + this.targetDirectory + "'" );
		LOG.info(LINE);
		
		Iterator<JsModule> moduleIt = this.modules.iterator();
		
		
		long start = System.currentTimeMillis();
		
		while (moduleIt.hasNext()){
			JsModule module = moduleIt.next();
			LOG.info("Processing module " + module.getName());
			processModule(module);
			LOG.info(LINE);
		}
		
		LOG.info("Done.");

		long end = System.currentTimeMillis();

		LOG.info("Build took " + Formatter.time(end - start));
				
		
	}
}
